@@ -148,6 +148,7 @@ struct nxt_http_request_s {
     nxt_str_t                       *path;
     nxt_str_t                       *args;
 
+    nxt_str_t                       args_decoded;
     nxt_array_t                     *arguments;  /* of nxt_http_name_value_t */
     nxt_array_t                     *cookies;    /* of nxt_http_name_value_t */
     nxt_list_t                      *fields;
@@ -226,9 +227,9 @@ struct nxt_http_action_s {
         nxt_upstream_t              *upstream;
         uint32_t                    upstream_number;
         nxt_var_t                   *var;
+        nxt_str_t                   *pass;
     } u;
 
-    nxt_str_t                       name;
     nxt_http_action_t               *fallback;
 };
 
@@ -313,7 +314,7 @@ nxt_int_t nxt_http_request_content_length(void *ctx, nxt_http_field_t *field,
 nxt_http_routes_t *nxt_http_routes_create(nxt_task_t *task,
     nxt_router_temp_conf_t *tmcf, nxt_conf_value_t *routes_conf);
 nxt_http_action_t *nxt_http_action_create(nxt_task_t *task,
-    nxt_router_temp_conf_t *tmcf, nxt_str_t *name);
+    nxt_router_temp_conf_t *tmcf, nxt_str_t *pass);
 nxt_int_t nxt_http_routes_resolve(nxt_task_t *task,
     nxt_router_temp_conf_t *tmcf);
 nxt_int_t nxt_http_pass_segments(nxt_mp_t *mp, nxt_str_t *pass,
